#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include "audacious/util.h"
#include "audacious/configdb.h"

#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <string.h>
#include <stdio.h>

#include <glib.h>
#include <glib/gi18n.h>

#include <gdk/gdkkeysyms.h>
#include <gtk/gtk.h>

#include "debug.h"
#include "config.h"
#include "configure.h"

static GtkWidget *entry_username = NULL, *entry_password = NULL;
static GtkWidget *chk_offline = NULL;
static GtkWidget *btn_apply = NULL, *btn_revert = NULL;
static GtkWidget *label_status = NULL;


GMutex *hatena_cfg_mutex;
hatena_configuration hatena_cfg;

static GMutex *m_status;


static void (* cb_update_cfg) (void);

static void
saveconfig(GtkWidget *wid __attribute__((unused)), gpointer data)
{
	ConfigDb *cfgfile;

	gchar *username = NULL, *password = NULL;
	gboolean offline = FALSE;

	username = (gchar *)gtk_entry_get_text(GTK_ENTRY(entry_username));
	password = (gchar *)gtk_entry_get_text(GTK_ENTRY(entry_password));

	offline = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(chk_offline));

	if ((cfgfile = bmp_cfg_db_open())) {
		bmp_cfg_db_set_string(cfgfile, "hatena", "username", username);
		bmp_cfg_db_set_string(cfgfile, "hatena", "password", password);
		bmp_cfg_db_set_bool(cfgfile, "hatena", "offline", offline);
		bmp_cfg_db_close(cfgfile);
	}


	//pdebug("mutex lock", DEBUG);
	g_mutex_lock(hatena_cfg_mutex);

	if (hatena_cfg.username) g_free(hatena_cfg.username);

	if (strlen(username) > 0)
		hatena_cfg.username = g_strdup(username);
	else
		hatena_cfg.username = NULL;

	if (hatena_cfg.password) g_free(hatena_cfg.password);

	if (strlen(password) > 0)
		hatena_cfg.password = g_strdup(password);
	else
		hatena_cfg.password = NULL;

	hatena_cfg.offline = offline;

	hatena_cfg.updated = TRUE;

	g_mutex_unlock(hatena_cfg_mutex);

	gtk_widget_set_sensitive(GTK_WIDGET(btn_apply), FALSE);
	gtk_widget_set_sensitive(GTK_WIDGET(btn_revert), FALSE);

	cb_update_cfg();
}

static void
resetconfig(GtkWidget *wid __attribute__((unused)), gpointer data)
{
	ConfigDb *db;

	if ((db = bmp_cfg_db_open())) {
		gchar *username = NULL, *password = NULL;
		bmp_cfg_db_get_string(db, "hatena", "username", &username);
		bmp_cfg_db_get_string(db, "hatena", "password", &password);

		if (hatena_cfg.username) g_free(hatena_cfg.username);

		if (username) {
			gtk_entry_set_text(GTK_ENTRY(entry_username), username);
			hatena_cfg.username = g_strdup(username);
			g_free(username);
		} else {
			hatena_cfg.username = NULL;
		}

		if (hatena_cfg.password) g_free(hatena_cfg.password);

		if (password) {
			gtk_entry_set_text(GTK_ENTRY(entry_password), password);
			hatena_cfg.password = g_strdup(password);
			g_free(password);
		} else {
			hatena_cfg.password = NULL;
		}


		hatena_cfg.offline = FALSE;
		
		bmp_cfg_db_get_bool(db, "hatena", "offline", 
							&hatena_cfg.offline);


		gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(chk_offline), 
									 hatena_cfg.offline);
		
		bmp_cfg_db_close(db);
	}

	gtk_widget_set_sensitive(GTK_WIDGET(btn_apply), FALSE);
	gtk_widget_set_sensitive(GTK_WIDGET(btn_revert), FALSE);
}


static void
modified(GtkWidget *wid __attribute__((unused)), gpointer data)
{
	gtk_widget_set_sensitive(GTK_WIDGET(btn_apply), TRUE);
	gtk_widget_set_sensitive(GTK_WIDGET(btn_revert), TRUE);
}


void
hatena_update_message(gchar *stat)
{

	//pdebug("mutex lock", DEBUG);
	g_mutex_lock(m_status);

	gtk_widget_hide (label_status);
	gtk_label_set_text(GTK_LABEL(label_status), stat);
	gtk_label_set_use_markup (GTK_LABEL (label_status), TRUE);
	gtk_widget_show (label_status);

	g_mutex_unlock(m_status);
}


/* Generated by glade, sorta. */
GtkWidget *
hatena_create_cfgdlg(void (* update_cb_func) (void))
{
	GtkWidget *vbox_top;

	GtkWidget *table_auth;
	GtkWidget *label_scrobbler;
	GtkWidget *label_username, *label_password;

	GtkWidget *himage1;
	GtkWidget *align_auth;
	GtkWidget *hbox;


	GtkWidget *frame_auth, *frame_connection, *frame_status;
	GtkWidget *vbox_connection;

	GtkWidget *align_connection, *align_status;

	cb_update_cfg = update_cb_func;
	hatena_cfg_mutex = g_mutex_new();

	m_status = g_mutex_new();

	vbox_top = gtk_vbox_new (FALSE, 0);

	label_scrobbler = gtk_label_new (_("<b>Hatena Preferences</b>"));
	gtk_widget_show (label_scrobbler);
	gtk_label_set_use_markup (GTK_LABEL (label_scrobbler), TRUE);
	gtk_misc_set_alignment (GTK_MISC (label_scrobbler), 0, 0.5);
	gtk_box_pack_start (GTK_BOX (vbox_top), label_scrobbler, FALSE, FALSE, 6);

	frame_auth = gtk_frame_new(_("Authentication"));
	gtk_widget_show (frame_auth);
	gtk_box_pack_start (GTK_BOX (vbox_top), frame_auth, FALSE, FALSE, 6);


	align_auth = gtk_alignment_new(0, 0, 0, 0);
	gtk_widget_show (align_auth);
	gtk_alignment_set_padding(GTK_ALIGNMENT(align_auth), 6, 12, 12, 0);

	table_auth = gtk_table_new (2, 2, FALSE);
	gtk_widget_show (table_auth);
	gtk_container_add(GTK_CONTAINER(align_auth), table_auth);

	gtk_container_add (GTK_CONTAINER (frame_auth), align_auth);

	gtk_table_set_row_spacings (GTK_TABLE(table_auth), 6);
	gtk_table_set_col_spacings (GTK_TABLE(table_auth), 6);

	label_username = gtk_label_new (_("Username:"));
	gtk_widget_show (label_username);
	gtk_table_attach_defaults (GTK_TABLE (table_auth), label_username, 0, 1, 0, 1);
	gtk_label_set_justify (GTK_LABEL (label_username), GTK_JUSTIFY_RIGHT);
	gtk_misc_set_alignment (GTK_MISC (label_username), 1, 0.5);

	label_password = gtk_label_new (_("Password:"));
	gtk_widget_show (label_password);
	gtk_table_attach (GTK_TABLE (table_auth), label_password, 0, 1, 1, 2,
					  (GtkAttachOptions) (GTK_FILL),
					  (GtkAttachOptions) (0), 0, 0);
	gtk_label_set_justify (GTK_LABEL (label_password), GTK_JUSTIFY_RIGHT);
	gtk_misc_set_alignment (GTK_MISC (label_password), 1, 0.5);

	entry_username = gtk_entry_new ();
	gtk_widget_show (entry_username);
	gtk_table_attach_defaults (GTK_TABLE (table_auth), entry_username, 1, 2, 0, 1);
	g_signal_connect(entry_username, "changed", (GCallback)modified, NULL);

	entry_password = gtk_entry_new ();
	gtk_entry_set_visibility(GTK_ENTRY(entry_password), FALSE);
	gtk_widget_show (entry_password);
	gtk_table_attach_defaults (GTK_TABLE (table_auth), entry_password, 1, 2, 1, 2);
	g_signal_connect(entry_password, "changed", (GCallback)modified, NULL);


	frame_connection = gtk_frame_new(_("Connection"));
	gtk_widget_show (frame_connection);
	gtk_box_pack_start (GTK_BOX (vbox_top), frame_connection, FALSE, FALSE, 6);

	align_connection = gtk_alignment_new(0, 0, 0, 0);
	gtk_widget_show (align_connection);
	gtk_alignment_set_padding(GTK_ALIGNMENT(align_connection), 6, 6, 12, 0);
	gtk_container_add (GTK_CONTAINER (frame_connection), align_connection);

	vbox_connection = gtk_vbox_new(FALSE, 0);
	gtk_widget_show (vbox_connection);
	gtk_container_add (GTK_CONTAINER (align_connection), vbox_connection);
	
	chk_offline = gtk_check_button_new_with_label(_("Offline"));
	gtk_widget_show (chk_offline);
	gtk_box_pack_start (GTK_BOX(vbox_connection), chk_offline, FALSE, FALSE, 0);
	g_signal_connect(chk_offline, "toggled", (GCallback)modified, NULL);



	frame_status = gtk_frame_new(_("Status"));
	gtk_widget_show (frame_status);
	gtk_box_pack_start (GTK_BOX (vbox_top), frame_status, FALSE, FALSE, 6);

	align_status = gtk_alignment_new(1.0, 0.5, 1.0, 1.0);
	gtk_widget_show (align_status);
	gtk_alignment_set_padding(GTK_ALIGNMENT(align_status), 6, 6, 6, 6);
	gtk_container_add (GTK_CONTAINER (frame_status), align_status);


	label_status = gtk_label_new ("");
	gtk_widget_show (label_status);
	gtk_label_set_use_markup (GTK_LABEL (label_status), TRUE);
	gtk_label_set_justify (GTK_LABEL (label_status), GTK_JUSTIFY_LEFT);
	gtk_label_set_line_wrap (GTK_LABEL(label_status), TRUE);
	gtk_misc_set_padding (GTK_MISC (label_status), 12, 0);
	gtk_misc_set_alignment (GTK_MISC (label_status), 0, 0.5);
	gtk_container_add (GTK_CONTAINER (align_status), label_status);



	himage1 
		= gtk_image_new_from_file (DATA_DIR "/images/hatena_badge.png");
	gtk_widget_show (himage1);
	gtk_box_pack_start (GTK_BOX (vbox_top), himage1, FALSE, FALSE, 0);
	gtk_misc_set_alignment (GTK_MISC (himage1), 1, 0.5);



	hbox = gtk_hbox_new(FALSE, 0);
	gtk_widget_show (hbox);

	btn_apply = gtk_button_new_from_stock(GTK_STOCK_APPLY);
	gtk_widget_show (btn_apply);

	gtk_widget_set_sensitive(GTK_WIDGET(btn_apply), FALSE);

	gtk_signal_connect(GTK_OBJECT(btn_apply), "clicked",
					   GTK_SIGNAL_FUNC(saveconfig), NULL);
	btn_revert = gtk_button_new_from_stock(GTK_STOCK_UNDO);
	gtk_widget_show (btn_revert);

	gtk_widget_set_sensitive(GTK_WIDGET(btn_revert), FALSE);

	gtk_signal_connect_object(GTK_OBJECT(btn_revert), "clicked",
							  GTK_SIGNAL_FUNC(resetconfig), NULL);
	gtk_box_pack_end(GTK_BOX(hbox), btn_revert, FALSE, FALSE, 3);
	gtk_box_pack_end(GTK_BOX(hbox), btn_apply, FALSE, FALSE, 3);

	gtk_box_pack_end(GTK_BOX(vbox_top), hbox, FALSE, FALSE, 0);




	gtk_entry_set_text(GTK_ENTRY(entry_username), "");
	gtk_entry_set_text(GTK_ENTRY(entry_password), "");

	resetconfig(NULL, NULL);


	return vbox_top;
}



