#include "config.h"

#include <glib.h>
#include <glib/gi18n.h>

#include <gdk/gdkkeysyms.h>
#include <gtk/gtk.h>

#include <curl/curl.h>

#include <audacious/plugin.h>
#include <audacious/prefswin.h>
#include <audacious/playlist.h>
#include <audacious/configdb.h>
#include <audacious/beepctrl.h>
#include <audacious/titlestring.h>

#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <wchar.h>
#include <sys/time.h>

#include "hatena.h"
#include "gtkstuff.h"
#include "config.h"
#include "debug.h"
#include "configure.h"

#define XS_CS xmms_hatena.xmms_session

#define XS_SLEEP 1

#define HS_SLEEP 10

#define AUDACIOUS_URL    "http://garakuta.homelinux.org/~nosuke/diary/"

typedef struct {
	int dosubmit;
	int pos_c;
	int len;
} submit_t;

static void init(void);
static void cleanup(void);
static void *xs_thread(void *);
static void *hs_thread(void *);

static GMutex *m_scrobbler;
static int going;

static GtkWidget *cfgdlg;

static GThread *pt_scrobbler;
static GThread *pt_handshake;

static GMutex *hs_mutex, *xs_mutex;
static GCond *hs_cond, *xs_cond;

static GMutex *m_status_handler;

static int initialized = 0;

static GeneralPlugin xmms_hatena =
{
	NULL,
	NULL,
	-1,
	NULL,
	init,
	hatena_about_show,
	NULL,
	cleanup
};



static void
configuration_update(void)
{
	g_cond_signal(hs_cond);
}



static void
status_update(sc_status_code status_code)
{

	GString *status = NULL, *message = NULL;

	CURLcode curl_code;
	time_t next;

	switch (status_code) {
	case NOT_RUNNING:
		status = g_string_new(_("Offline"));
		if (hatena_cfg.offline)
			message = g_string_new(_("Offline mode"));
		else if (! hatena_cfg.username)
			message = g_string_new(_("Username is empty"));
		else if (! hatena_cfg.password)
			message = g_string_new(_("Password is empty"));
		else 
			message = g_string_new(_("Unknown reason"));
		break;

	case HS_EMPTY_BODY:
	case HS_SERVER_FAILED:
		/*case HS_NOINTERVAL:*/
	case HS_BROKEN_RESPONSE:
	case HS_UNKNOWN_ERROR:
		status = g_string_new(_("Bad"));
		next = hatena_get_next_hs();
		message = g_string_new("");
		g_string_printf(message, _("Handshake Failed (%d).  Retry at : %s"),
						status_code, ctime(&next));

		break;

	case HS_UPDATE_PLUGIN:
		status = g_string_new(_("Hanging Up"));
		message = g_string_new(_("Update Audacious Hatena plugin."));

		//g_mutex_lock(m_status_handler);
		hatena_errorbox_show(ht_fmt_vastr(_("Please update Audacious Hatena plugin.\n"
											   "Update available at: %s"), 
											 AUDACIOUS_URL));
		//g_mutex_unlock(m_status_handler);		

		break;

	case HS_OK:
		status = g_string_new(_("Fine"));
		message = g_string_new("Handshake OK.");
		break;

	case HS_BADUSERNAME:
		status = g_string_new(_("Hanging Up"));
		message = g_string_new(_("Bad user name."));

		//g_mutex_lock(m_status_handler);
		hatena_errorbox_show(_("Incorrect username/password.\n"
								  "Please fix in configuration."));
		//g_mutex_unlock(m_status_handler);		

		break;

	case HS_NOUSERNAME:
		status = g_string_new(_("Hanging Up"));
		message = g_string_new(_("Set your user ID."));
		break;

	case HS_NOPASSWORD:
		status = g_string_new(_("Hanging Up"));
		message = g_string_new(_("Set your password."));
		break;

	case HS_CURL_CONNECT:
		status = g_string_new(_("Connecting"));
		message = g_string_new("Handshaking.");
		break;

	case HS_CURL_ERROR:
		status = g_string_new(_("Bad"));
		next = hatena_get_next_hs();
		curl_code = hatena_get_curl_status();
		message = g_string_new("");
		g_string_printf(message,
						_("Handshake Failed (cURL error).  cURL error code: %2d"), curl_code);

		break;

	case SB_EMPTY_BODY:
	case SB_SERVER_FAILED:
	case SB_UNKNOWN_ERROR:
	case SB_BADAUTHENTICATON_1ST:
		status = g_string_new(_("Bad"));
		next = hatena_get_next_sb();
		message = g_string_new("");
		g_string_printf(message, _("Submission Failed (%d).  Retry at : %s"),
						status_code, ctime(&next));

		break;

	case SB_OK:
		status = g_string_new(_("Fine"));
		message = g_string_new(_("Submission OK."));
		break;

	case SB_BADAUTHENTICATON:
		status = g_string_new(_("Hanging Up"));
		message = g_string_new(_("Bad username or password."));

		//g_mutex_lock(m_status_handler);
		hatena_errorbox_show(_("Incorrect username/password.\n"
								  "Please fix in configuration."));
		//g_mutex_unlock(m_status_handler);		

		break;


	case SB_CURL_CONNECT:
		status = g_string_new(_("Connecting"));
		message = g_string_new(_("Submitting"));
		break;

	case SB_CURL_ERROR:
		status = g_string_new(_("Bad"));
		next = hatena_get_next_sb();
		curl_code = hatena_get_curl_status();
		message = g_string_new("");
		g_string_printf(message, 
						_("Submission Failed (cURL error).  cURL error code: %2d"), curl_code);

		break;

	default:
		status = g_string_new(_("Unknown"));
		message = g_string_new(_("Unknown state."));
		break;
	}

	if (status->str) {
		GString *msg;
		msg = g_string_new("");
		g_string_printf(msg, "<b>%s</b>\n%s\n", status->str, 
						message->str ? message->str : "");
		
		hatena_update_message(msg->str);

		g_string_free(msg, TRUE);
	}

	if (status)
		g_string_free(status, TRUE);

	if (message)
		g_string_free(message, TRUE);

}

static void
status_update_active(void)
{
	sc_status_code status_code;

	status_code = hatena_get_status();

	status_update(status_code);
	
}


/* Don't call this function from main thread */
static void
status_update_handler(sc_status_code status_code)
{
	gdk_threads_enter();

	status_update(status_code);

	gdk_threads_leave();
}


static void
init(void)
{
	going = 1;
	GError **moo = NULL;
	gchar buf[PATH_MAX];
	const gchar *home;
	//sc_status_code status;
	//CURLcode curl_status;
	ConfigDb *db;

	m_scrobbler = g_mutex_new();
	hs_mutex = g_mutex_new();
	xs_mutex = g_mutex_new();
	hs_cond = g_cond_new();
	xs_cond = g_cond_new();

	m_status_handler = g_mutex_new();

	if ((db = bmp_cfg_db_open())) {
		// read at initialization stage only
		hatena_cfg.useproxy = FALSE;
		hatena_cfg.useproxyauth = FALSE;
		hatena_cfg.proxy_server = NULL;
		hatena_cfg.proxy_port = NULL;
		hatena_cfg.proxy_username = NULL;
		hatena_cfg.proxy_password = NULL;

		bmp_cfg_db_get_bool(db, NULL, "use_proxy",
							&hatena_cfg.useproxy);

		if (hatena_cfg.useproxy) {
			bmp_cfg_db_get_string(db, NULL, "proxy_host", 
								  &hatena_cfg.proxy_server);
			bmp_cfg_db_get_string(db, NULL, "proxy_port", 
								  &hatena_cfg.proxy_port);
		
			if (! hatena_cfg.proxy_server) {
				hatena_cfg.useproxy = FALSE;
				hatena_cfg.proxy_port = NULL;
			} else {
				bmp_cfg_db_get_bool(db, NULL, "proxy_use_auth", 
									&hatena_cfg.useproxyauth);
				if (hatena_cfg.useproxyauth) {
					bmp_cfg_db_get_string(db, NULL, "proxy_user", 
										  &hatena_cfg.proxy_username);
					bmp_cfg_db_get_string(db, NULL, "proxy_pass", 
										  &hatena_cfg.proxy_password);
				}
				
			}
		}
		bmp_cfg_db_close(db);
	}


	cfgdlg = hatena_create_cfgdlg(configuration_update);
	
	prefswin_page_new(cfgdlg, "Hatena", DATA_DIR "/images/hatena.png");

	hatena_init();

#define HATENA_CACHE ".audacious/hatenaqueue.txt"
	if ((home = g_getenv("HOME")) || (home = g_get_home_dir())) {
		if (g_snprintf(buf, PATH_MAX, "%s/" HATENA_CACHE, home) < PATH_MAX)
			hatena_init_cache(buf);	
		else
			pdebug(ht_fmt_vastr("home directory not found: %s", home), DEBUG);
	}

	hatena_set_status_handler(status_update_handler);

	status_update_active();
	
	hatena_set_username(hatena_cfg.username);
	hatena_set_password(hatena_cfg.password);
	hatena_set_proxy(hatena_cfg.useproxy, hatena_cfg.proxy_server, 
						hatena_cfg.proxy_port, hatena_cfg.useproxyauth, 
						hatena_cfg.proxy_username, 
						hatena_cfg.proxy_password);

	if ((pt_scrobbler = g_thread_create(xs_thread, m_scrobbler, TRUE, moo)) 
		== NULL) {
		pdebug(ht_fmt_vastr("Error creating hatena thread: %s", moo), DEBUG);
		return;
	}


	if ((pt_handshake = g_thread_create(hs_thread, m_scrobbler, TRUE, NULL)) 
		== NULL) {
		pdebug("Error creating handshake thread", DEBUG);
		return;
	}

	pdebug("plugin started", DEBUG);

	initialized = 1;
}


static void
cleanup(void)
{
	if (! initialized)
		return;
	
	if (xmms_hatena.description)
		g_free(xmms_hatena.description);
    xmms_hatena.description = NULL;

	pdebug("about to lock mutex", DEBUG);
	//pdebug("mutex lock", DEBUG);
	g_mutex_lock(m_scrobbler);
	pdebug("locked mutex", DEBUG);
	going = -1;
	g_mutex_unlock(m_scrobbler);

	pdebug("joining threads", DEBUG);

	pdebug("send signal to xs", DEBUG);

	//pdebug("mutex lock", DEBUG);
	g_mutex_lock(xs_mutex);
	g_cond_signal(xs_cond);
	g_mutex_unlock(xs_mutex);

	pdebug("send signal to hs", DEBUG);

	//pdebug("mutex lock", DEBUG);
	g_mutex_lock(hs_mutex);
	g_cond_signal(hs_cond);
	g_mutex_unlock(hs_mutex);

	pdebug("wait xs", DEBUG);

	g_thread_join(pt_scrobbler);

	pdebug("wait hs", DEBUG);

	g_thread_join(pt_handshake);

	g_cond_free(hs_cond);
	g_cond_free(xs_cond);
	g_mutex_free(hs_mutex);
	g_mutex_free(xs_mutex);

	pdebug("page destroy", DEBUG);

	prefswin_page_destroy(cfgdlg);

	pdebug("sc cleaner", DEBUG);

	hatena_cleaner();

	initialized  = 0;
}


static char
ishttp(const char *a)
{
	char *tmp, *bp;
	int status = 0;

	if (!a || !*a)
		return 0;

	tmp = strdup(a);
	for (bp = tmp; *bp; bp++)
		*bp = toupper((int) *bp);
	if (strstr(tmp, "HTTP://"))
		status = -1;
	free(tmp);
	return status;
}



/* Following code thanks to nosuke 
 *
 * It should probably be cleaned up
 */
static submit_t
get_song_status(void)
{
	static int pos_c, playlistlen_c, playtime_c, time_c,
		pos_p = 0, playlistlen_p = 0, playtime_p = 0,
		playtime_i = 0, time_i = 0,
		playtime_ofs = 0;
	static char *file_c = NULL, *file_p = NULL; 

	static enum playstatus {
		ps_stop, ps_play, ps_pause
	} ps_c, ps_p = ps_stop;

	static int submitted = 0, changed, seeked, repeat,
		filechanged, rewind, len = 0;

	static enum state {
		start, stop, pause, restart, playing, pausing, stopping
	} playstate;

	submit_t dosubmit;

	struct timeval timetmp;

	/* clear dosubmit */
	dosubmit.dosubmit = dosubmit.pos_c = dosubmit.len = 0;

	/* current music number */
	pos_c = xmms_remote_get_playlist_pos(XS_CS);
	/* current file name */
	file_c = xmms_remote_get_playlist_file(XS_CS, pos_c); 
	/* total number */
	playlistlen_c = xmms_remote_get_playlist_length(XS_CS);
	/* current playtime */
	playtime_c = xmms_remote_get_output_time(XS_CS); 
	/* total length */
	len = xmms_remote_get_playlist_time(XS_CS, pos_c); 

	/* current time (ms) */
	gettimeofday(&timetmp, NULL);
	time_c = timetmp.tv_sec * 1000 + timetmp.tv_usec / 1000; 

	/* current status */
	if( xmms_remote_is_paused(XS_CS) ) {
		ps_c = ps_pause;
	}else if( xmms_remote_is_playing(XS_CS) ) {
		ps_c = ps_play;
	}else{
		ps_c = ps_stop;
	}

	/* repeat setting */
	repeat = xmms_remote_is_repeat(XS_CS);


	if (ps_p == ps_stop && ps_c == ps_stop)         playstate = stopping;
	else if (ps_p == ps_stop && ps_c == ps_play)    playstate = start;
	else if (ps_p == ps_play && ps_c == ps_play)    playstate = playing;
	else if (ps_p == ps_play && ps_c == ps_stop)    playstate = stop;
	else if (ps_p == ps_play && ps_c == ps_pause)   playstate = pause;
	else if (ps_p == ps_pause && ps_c == ps_pause)  playstate = pausing;
	else if (ps_p == ps_pause && ps_c == ps_play)   playstate = restart;
	else if (ps_p == ps_pause && ps_c == ps_stop)   playstate = stop;
	else playstate = stopping;

	/* filename has changed */
	if (! (file_p == NULL && file_c == NULL)
		&& ((file_p == NULL && file_c != NULL)
			|| (file_p != NULL && file_c == NULL)
			|| (file_p != NULL && file_c != NULL && strcmp(file_c, file_p)))) {
		filechanged = 1;
		pdebug("*** filechange ***", SUB_DEBUG);
	}else{
		filechanged = 0;
	}

	if (file_c == NULL) 
		len = 0;

	/* whole rewind has occurred (maybe) */
	if (len != 0 && len - (playtime_p - playtime_c) < 3000) {
		rewind = 1;
		pdebug("*** rewind ***", SUB_DEBUG);
	} else {
		rewind = 0;
	}

	changed = 0;
	seeked = 0;

	switch (playstate) {
	case start:
	  pdebug("*** START ***", SUB_DEBUG);
	  break;
	case stop:
	  pdebug("*** STOP ***", SUB_DEBUG);
	  len = 0;
	  break;
	case pause: 
	  pdebug("*** PAUSE ***", SUB_DEBUG);
	  playtime_ofs += playtime_c - playtime_i; /* save playtime */
	  break;
	case restart: 
	  pdebug("*** RESTART ***", SUB_DEBUG);
	  playtime_i  = playtime_c; /* restore playtime */
	  break;
	case playing:
	  if ((playtime_c < playtime_p) /* back */
		  || ((playtime_c - playtime_i) 
			  - (time_c - time_i) > 3000) /* forward */)
		  seeked = 1;

	  if (filechanged || /* filename has changed */
		  (! filechanged /* filename has not changed... */
		   /* looping with only one file */
		   && ((pos_c == 0 && playlistlen_c == 1 && repeat && rewind)
			   || (pos_p == pos_c && rewind) /* looping? */ 
			   || (pos_p != pos_c && seeked) 
			   /* skip from current music to next music, 
				  which has the same filename as previous one */
			   ||( pos_p < pos_c && playtime_c < playtime_p ) || 
			   /* current song has removed from playlist 
				  but the next (following) song has the same
				  filename */
			   (playlistlen_p > playlistlen_c && playtime_c < playtime_p)))) {
		  pdebug("*** CHANGE ***",SUB_DEBUG);
		  pdebug(ht_fmt_vastr(" filechanged = %d",filechanged),SUB_DEBUG);
		  pdebug(ht_fmt_vastr(" pos_c = %d",pos_c),SUB_DEBUG);
		  pdebug(ht_fmt_vastr(" pos_p = %d",pos_p),SUB_DEBUG);
		  pdebug(ht_fmt_vastr(" rewind = %d", rewind),SUB_DEBUG);
		  pdebug(ht_fmt_vastr(" seeked = %d", seeked),SUB_DEBUG);
		  pdebug(ht_fmt_vastr(" playtime_c = %d", playtime_c),SUB_DEBUG);
		  pdebug(ht_fmt_vastr(" playtime_p = %d", playtime_p),SUB_DEBUG);
		  pdebug(ht_fmt_vastr(" playlistlen_c = %d", playlistlen_p), SUB_DEBUG);
		  pdebug(ht_fmt_vastr(" playlistlen_p = %d", playlistlen_p), SUB_DEBUG);
		  changed = 1;
		  seeked = 0;

		  if (file_p != NULL) {
			  g_free(file_p);
			  file_p = NULL;
		  }

	  } else if (seeked) { 
		  seeked = 1;
		  pdebug("*** SEEK ***", SUB_DEBUG);
	  }
	  break;
	case pausing: 
		if (playtime_c != playtime_p) {
			pdebug("*** SEEK ***", SUB_DEBUG);
			seeked = 1;
		}
		break;
	case stopping:
		len = 0;
		break;
	default:
		pdebug("*** unknown state tranfer!!! ***", SUB_DEBUG);
		break;
	}

	
	if (playstate == start || changed || (seeked && !submitted)) {
	  /* reset counter */
	  pdebug(" <<< reset counter >>>", SUB_DEBUG);

	  submitted = 0;
	  playtime_ofs = 0;
	  playtime_i = playtime_c;
	  time_i = time_c;

	} else {
	  /* check playtime for submitting */
	  if (!submitted) {
		if (len > 30 * 1000 
			&& ((playtime_ofs + playtime_c - playtime_i > len / 2) 
#if 0
				// DEBUG
				|| (playtime_ofs + playtime_c - playtime_i > 3 * 1000)
#endif 
				|| (playtime_ofs + playtime_c - playtime_i > 240 * 1000))) {
			pdebug("*** submitting requirements are satisfied.", SUB_DEBUG);
			pdebug(ht_fmt_vastr("    len = %d, playtime = %d",
							 len / 1000, (playtime_c - playtime_i)/1000 ),
				   SUB_DEBUG);
			submitted = 1;
			dosubmit.dosubmit = 1;
			dosubmit.pos_c = pos_c;
			dosubmit.len = len;
		}
	  }
	}

	g_free(file_p);

	/* keep current value for next iteration */
	ps_p = ps_c;
	file_p = file_c;
	playtime_p = playtime_c;
	pos_p = pos_c;
	playlistlen_p = playlistlen_c;

	return dosubmit;
}





static void *
xs_thread(void *data __attribute__((unused)))
{
	int run = 0;
	submit_t dosubmit;
	GTimeVal sleeptime;
	
	while (1) {

		//pdebug("mutex lock", DEBUG);
		g_mutex_lock(m_scrobbler);
		run = going;
		g_mutex_unlock(m_scrobbler);

		if (run < 0)
			break;

		
		/* Check for ability to submit */
		dosubmit = get_song_status();
		
		if(dosubmit.dosubmit) {
			TitleInput *tuple;
				
			pdebug("Submitting song.", DEBUG);

			tuple = playlist_get_tuple(dosubmit.pos_c);

			if (ishttp(tuple->file_name))
				continue;

			if (tuple->performer && tuple->track_name ) {
				//char utctime[20];
				//time_t tm = time(NULL);

				pdebug(ht_fmt_vastr("submitting artist: %s, title: %s",
								 tuple->performer, tuple->track_name),
					   DEBUG);
				//hatena_add_entry(tuple, dosubmit.len/1000);
				//strftime(utctime, sizeof(utctime), "%F %T", gmtime(&tm));
				
				hatena_add_entry(tuple->performer, tuple->track_name, 
									//utctime, tuple->album_name,
									time(NULL), tuple->album_name,
#ifdef NOTYET
									tuple->mb,
#else
									NULL,
#endif
									dosubmit.len/1000);
				
			} else {
				pdebug("tuple does not contain an artist or a title,"
					   "not submitting.", DEBUG);
			}
		}

		g_get_current_time(&sleeptime);

		sleeptime.tv_sec += XS_SLEEP;

		//pdebug("xs wait", DEBUG);
		//pdebug("mutex lock", DEBUG);
		g_mutex_lock(xs_mutex);
		g_cond_timed_wait(xs_cond, xs_mutex, &sleeptime);
		g_mutex_unlock(xs_mutex);

	}

	pdebug("hatena thread: exiting", DEBUG);
	g_thread_exit(NULL);

	return NULL;
}


static void *
hs_thread(void *data __attribute__((unused)))
{
	int run = 1;
	GTimeVal sleeptime;


	pdebug("hs_thread start", DEBUG);

	while (1) {

		//pdebug("mutex lock", DEBUG);
		g_mutex_lock(m_scrobbler);
		run = going;
		g_mutex_unlock(m_scrobbler);

		if (run < 0) // exit
			break;

		//pdebug("mutex lock", DEBUG);
		g_mutex_lock(hatena_cfg_mutex);


		if (hatena_cfg.updated) {
			pdebug("config file updated", DEBUG);
			hatena_reset_params();
			hatena_set_username(hatena_cfg.username);
			hatena_set_password(hatena_cfg.password);

			hatena_set_proxy(hatena_cfg.useproxy, 
								hatena_cfg.proxy_server, 
								hatena_cfg.proxy_port,
								hatena_cfg.useproxyauth,
								hatena_cfg.proxy_username,
								hatena_cfg.proxy_password);

			hatena_cfg.updated = 0;
		}

		//pdebug("mutex lock2", DEBUG);
		g_mutex_lock(m_scrobbler);


		if (going >= 0) {
			if (hatena_cfg.offline 
				|| ! hatena_cfg.username 
				|| ! hatena_cfg.password) {
				going = 0;
			} else
				going = 1;

			if (! going)
				status_update_handler(NOT_RUNNING);
		}
		run = going;

		g_mutex_unlock(m_scrobbler);

		g_mutex_unlock(hatena_cfg_mutex);

		if (run < 0) 
			break;

		if (run) {
			if (hatena_idle() < 0) {
				pdebug("Giving up due to fatal error", DEBUG);
				//pdebug("mutex lock", DEBUG);
				g_mutex_lock(m_scrobbler);
				going = 0;
				g_mutex_unlock(m_scrobbler);
				run = going;
			}
			status_update_active();
		}
		

		g_get_current_time(&sleeptime);

		if (run)
			sleeptime.tv_sec += HS_SLEEP;

		pdebug("hs wait", DEBUG);
		//pdebug("mutex lock", DEBUG);
		g_mutex_lock(hs_mutex);
		if (run)
			g_cond_timed_wait(hs_cond, hs_mutex, &sleeptime);
		else
			g_cond_wait(hs_cond, hs_mutex);

		g_mutex_unlock(hs_mutex);

	}

	pdebug("handshake thread: exiting", DEBUG);
	g_thread_exit(NULL);

	return NULL;
}



GeneralPlugin *
get_gplugin_info(void) {
	xmms_hatena.description = g_strdup_printf(_("Hatena Plugin"));
	return &xmms_hatena;
}
